//+------------------------------------------------------------------+
//|                                                     MapArray.mqh |
//|                    Copyright 2021-2025, MetaQuotes Software Ltd. |
//|                                             https://www.mql5.com |
//+------------------------------------------------------------------+

//+------------------------------------------------------------------+
//| Generic pair [key;value]                                         |
//+------------------------------------------------------------------+
template<typename K,typename V>
struct Pair
{
   K key;
   V value;
   
   bool operator==(const Pair<K,V> &second) const
   {
      return key == second.key;
   }

   bool operator>(const Pair<K,V> &second) const
   {
      return key > second.key;
   }
};

//+------------------------------------------------------------------+
//| Array of pairs [key;value]                                       |
//+------------------------------------------------------------------+
template<typename K,typename V>
class MapArray
{
protected:
   Pair<K,V> array[];
   int keyindex[]; // slim index array for sorting and binary searching by keys

   // array is stored 'as is', order by index
   template<typename T>
   int findOrInsertByRef(const T &value, const bool insert = true)
   {
      ResetLastError();
      const int n = ArraySize(array);
      int i = 0, min = 0, max = n - 1;
      while(max >= min)
      {
         i = (max + min) / 2;
         if(array[keyindex[i]] == value) return keyindex[i];
         if(value > array[keyindex[i]]) min = fmax(i, min + 1);
         else max = fmin(i, max - 1);
      }
      
      SetUserError(1); // not found
      
      // min > max
      if(min >= n)
      {
         if(!insert) return n;
         ArrayResize(array, n + 1, 2 * n);
         ArrayResize(keyindex, n + 1, 2 * n);
         array[n] = value;
         keyindex[n] = n;
         return n;
      }
      else if(max < 0)
      {
         if(!insert) return 0;
         ArrayResize(array, n + 1, 2 * n);
         ArraySetAsSeries(keyindex, true);
         ArrayResize(keyindex, n + 1, 2 * n);
         array[n] = value;
         keyindex[n] = n; // means keyindex[0] = n;
         ArraySetAsSeries(keyindex, false);
         return n; // always added to the end of array
      }
      
      if(!insert) return keyindex[i];
   
      ArrayResize(array, n + 1, 2 * n);
      ArrayResize(keyindex, n + 1, 2 * n);
      
      array[n] = value;
      
      ArraySetAsSeries(keyindex, true);
      ArrayCopy(keyindex, keyindex, 0, 1, n - i);
      ArraySetAsSeries(keyindex, false);
      /* equivalent
      for(int j = n - 1; j >= i; j--)
      {
         keyindex[j + 1] = keyindex[j];
      }
      */
      
      keyindex[i] = n;
      return n;
   }

   template<typename T>
   int findByRef(const T &value) const
   {
      ResetLastError();
      const int n = ArraySize(array);
      int i = 0, min = 0, max = n - 1;
      while(max >= min)
      {
         i = (max + min) / 2;
         if(array[keyindex[i]] == value) return keyindex[i];
         if(value > array[keyindex[i]]) min = fmax(i, min + 1);
         else max = fmin(i, max - 1);
      }
      
      SetUserError(1); // not found
      
      // min > max
      if(min >= n)
      {
         return n;
      }
      else if(max < 0)
      {
         return 0;
      }
      
      return keyindex[i];
   }

   int add(const K k, const V v)
   {
      const int n = ArraySize(array);
      Pair<K,V> p = {k, v};
      findOrInsertByRef(p);
      return n;
   }

public:
   int getSize() const
   {
      return ArraySize(array);
   }
   
   K getKey(const int i, const bool sorted = false) const
   {
      if(i < 0 || i >= ArraySize(array))
      {
      #ifdef _DEBUG
         DebugBreak();
      #endif
         return NULL;
      }
      return array[sorted ? keyindex[i] : i].key;
   }
   
   V getValue(const int i, const bool sorted = false) const
   {
      if(i < 0 || i >= ArraySize(array))
      {
      #ifdef _DEBUG
         DebugBreak();
      #endif
         return NULL;
      }
      return array[sorted ? keyindex[i] : i].value;
   }
   
   int find(const K k) const
   {
      
      Pair<K,V> p = {k, NULL};
      int i = findByRef(p);
      if(!_LastError) return i;
      
      /* equivalent
      for(int i = 0; i < ArraySize(array); ++i)
      {
         if(array[i].key == k)
         {
            return i;
         }
      }
      */
      return -1;
   }
   
   int put(const K k, const V v)
   {
      Pair<K,V> p = {k, v};
      const int i = findOrInsertByRef(p);
      if(!_LastError)
      {
         array[i].value = v;
      }
      return i;
   }

   int inc(const K k, const V plus = (V)1)
   {
      const int i = find(k);
      if(i != -1)
      {
         array[i].value += plus;
         return i;
      }
      else return add(k, plus);
   }
   
   void remove(const K k)
   {
      const int i = find(k);
      if(i != -1)
      {
         for(int j = i + 1; j < ArraySize(array); ++j)
         {
            array[j - 1] = array[j];
         }

         ArrayResize(array, ArraySize(array) - 1);
         
         // shrink keyindex[], decrement all elements larger than [i]
         int k = -1;
         for(int j = 0; j < ArraySize(keyindex); j++)
         {
            if(keyindex[j] == i)
            {
               k = j;
               break;
            }
         }
         
         for(int j = k + 1; j < ArraySize(keyindex); ++j)
         {
            keyindex[j - 1] = keyindex[j];
         }
         
         ArrayResize(keyindex, ArraySize(keyindex) - 1);
         for(int j = 0; j < ArraySize(keyindex); j++)
         {
            if(keyindex[j] > i) keyindex[j]--;
         }
      }
   }
   
   Pair<K,V> operator[](const int i) // original order is preserved
   {
      if(i < 0 || i >= ArraySize(array))
      {
         static Pair<K,V> null = {};
      #ifdef _DEBUG
         DebugBreak();
      #endif
         return null;
      }
      return array[i];
   }

   V operator[](const K k)
   {
      const int i = find(k);
      if(i != -1) return array[i].value;
      return (V)NULL;
   }

   void reset()
   {
      ArrayResize(array, 0);
   }
   
   void print() const
   {
      ArrayPrint(array);
   }
};
//+------------------------------------------------------------------+
